# For Developers

Developers may fork [the GTPSA.jl repo](https://github.com/bmad-sim/GTPSA.jl) and then `dev` their forked repo in the REPL. For example, if my Github username is `githubuser`, then after forking GTPSA.jl I would run in the REPL:

```julia
import Pkg
Pkg.develop(url="https://github.com/githubuser/GTPSA.jl")
```

The package consists of two layers: a low-level layer written in Julia that is 1-to-1 with the GTPSA C code, and a high-level, user-friendly layer that cleans up the notation for manipulating TPSAs, manages temporaries generated during evaluation, and properly manages the memory in C when variables go out of scope in Julia. The low-level functions, which are exported for developer usage at the moment, are listed below.

The C code consists of three C structs: `desc`, `tpsa`, and `ctpsa`. The low-level Julia-equivalent, 1-to-1 structs are respectively `Desc`, `RTPSA`, and `CTPSA`. C pointers `Ptr` to these structs are wrapped by the high-level structs `Descriptor`, `TPSA`, and `ComplexTPSA` respectively.

The low-level structs `Desc`, `RTPSA`, and `CTPSA` contain fields with `Ptr{Cvoid}`; these are pointers to any of the other structs, specified in the documentation. For example, the `d` field in `RTPSA` is a `Ptr{Desc}`. We could not explicitly define `Ptr{Desc}`, because `Desc` likewise has a `Ptr{RTPSA}`, and Julia does not allow for cyclic implicit struct definitions. Therefore, the `Ptr` must be converted to the appropriate low-level Julia struct before safe accessing (documented next to each of the fields).

For example, to access the `t` array of `RTPSA`s in `Desc` defined by a high-level struct `Descriptor`:

```
using GTPSA
import GTPSA: RTPSA, CTPSA, Desc

d = Descriptor(5,8)

# To access the low-level C struct:
desc = unsafe_load(d.desc)

# To access the array of 8 temporaries t:
t_jl = unsafe_wrap(Vector{Ptr{RTPSA}}, Base.unsafe_convert(Ptr{Ptr{RTPSA}}, desc.t), 8)
```

First the `Ptr{Ptr{Cvoid}}` is converted to a `Ptr{Ptr{RTPSA}}`, which can then be wrapped in a `Vector{Ptr{RTPSA}}`. Note that calling `unsafe_wrap` uses memory allocation, so it is preferred to use `unsafe_load` for speed-critical applications.To get the first temporary in the array without using `unsafe_wrap`, use:

```
t1_jl = unsafe_load(Base.unsafe_convert(Ptr{Ptr{RTPSA}}, desc.t), 1)
```

## Monomial
```@docs
GTPSA.mad_mono_str!
GTPSA.mad_mono_prt!
GTPSA.mad_mono_fill!
GTPSA.mad_mono_copy!
GTPSA.mad_mono_min
GTPSA.mad_mono_max
GTPSA.mad_mono_ord
GTPSA.mad_mono_ordp
GTPSA.mad_mono_ordpf
GTPSA.mad_mono_eq
GTPSA.mad_mono_lt
GTPSA.mad_mono_le
GTPSA.mad_mono_cmp
GTPSA.mad_mono_rcmp
GTPSA.mad_mono_add!
GTPSA.mad_mono_sub!
GTPSA.mad_mono_cat!
GTPSA.mad_mono_rev!
GTPSA.mad_mono_print
```

## Desc
```@docs
GTPSA.Desc
GTPSA.mad_desc_newv
GTPSA.mad_desc_newvp
GTPSA.mad_desc_newvpo
GTPSA.mad_desc_del!
GTPSA.mad_desc_getnv!
GTPSA.mad_desc_maxord
GTPSA.mad_desc_maxlen
GTPSA.mad_desc_gtrunc!
GTPSA.mad_desc_isvalids
GTPSA.mad_desc_isvalidm
GTPSA.mad_desc_isvalidsm
GTPSA.mad_desc_idxs
GTPSA.mad_desc_idxm
GTPSA.mad_desc_idxsm
GTPSA.mad_desc_nxtbyvar
GTPSA.mad_desc_nxtbyord
GTPSA.mad_desc_mono!
GTPSA.mad_desc_info
```
## RTPSA
```@docs
GTPSA.RTPSA
GTPSA.mad_tpsa_newd
GTPSA.mad_tpsa_new
GTPSA.mad_tpsa_del!
GTPSA.mad_tpsa_desc
GTPSA.mad_tpsa_uid!
GTPSA.mad_tpsa_len
GTPSA.mad_tpsa_nam
GTPSA.mad_tpsa_ord
GTPSA.mad_tpsa_ordv
GTPSA.mad_tpsa_ordn
GTPSA.mad_tpsa_copy!
GTPSA.mad_tpsa_sclord!
GTPSA.mad_tpsa_getord!
GTPSA.mad_tpsa_cutord!
GTPSA.mad_tpsa_maxord!
GTPSA.mad_tpsa_convert!
GTPSA.mad_tpsa_setvar!
GTPSA.mad_tpsa_setval!
GTPSA.mad_tpsa_setnam!
GTPSA.mad_tpsa_clear!
GTPSA.mad_tpsa_isnul
GTPSA.mad_tpsa_mono!
GTPSA.mad_tpsa_idxs
GTPSA.mad_tpsa_idxm
GTPSA.mad_tpsa_idxsm
GTPSA.mad_tpsa_cycle!
GTPSA.mad_tpsa_get0
GTPSA.mad_tpsa_geti
GTPSA.mad_tpsa_gets
GTPSA.mad_tpsa_getm
GTPSA.mad_tpsa_getsm
GTPSA.mad_tpsa_set0!
GTPSA.mad_tpsa_seti!
GTPSA.mad_tpsa_sets!
GTPSA.mad_tpsa_setm!
GTPSA.mad_tpsa_setsm!
GTPSA.mad_tpsa_getv!
GTPSA.mad_tpsa_setv!
GTPSA.mad_tpsa_equ
GTPSA.mad_tpsa_dif!
GTPSA.mad_tpsa_add!
GTPSA.mad_tpsa_sub!
GTPSA.mad_tpsa_mul!
GTPSA.mad_tpsa_div!
GTPSA.mad_tpsa_pow!
GTPSA.mad_tpsa_powi!
GTPSA.mad_tpsa_pown!
GTPSA.mad_tpsa_nrm
GTPSA.mad_tpsa_abs!
GTPSA.mad_tpsa_sqrt!
GTPSA.mad_tpsa_exp!
GTPSA.mad_tpsa_log!
GTPSA.mad_tpsa_sincos!
GTPSA.mad_tpsa_sin!
GTPSA.mad_tpsa_cos!
GTPSA.mad_tpsa_tan!
GTPSA.mad_tpsa_cot!
GTPSA.mad_tpsa_sinc!
GTPSA.mad_tpsa_sincosh!
GTPSA.mad_tpsa_sinh!
GTPSA.mad_tpsa_cosh!
GTPSA.mad_tpsa_tanh!
GTPSA.mad_tpsa_coth!
GTPSA.mad_tpsa_sinhc!
GTPSA.mad_tpsa_asin!
GTPSA.mad_tpsa_acos!
GTPSA.mad_tpsa_atan!
GTPSA.mad_tpsa_acot!
GTPSA.mad_tpsa_asinc!
GTPSA.mad_tpsa_asinh!
GTPSA.mad_tpsa_acosh!
GTPSA.mad_tpsa_atanh!
GTPSA.mad_tpsa_acoth!
GTPSA.mad_tpsa_asinhc!
GTPSA.mad_tpsa_erf!
GTPSA.mad_tpsa_erfc!
GTPSA.mad_tpsa_acc!
GTPSA.mad_tpsa_scl!
GTPSA.mad_tpsa_inv!
GTPSA.mad_tpsa_invsqrt!
GTPSA.mad_tpsa_unit!
GTPSA.mad_tpsa_atan2!
GTPSA.mad_tpsa_hypot!
GTPSA.mad_tpsa_hypot3!
GTPSA.mad_tpsa_integ!
GTPSA.mad_tpsa_deriv!
GTPSA.mad_tpsa_derivm!
GTPSA.mad_tpsa_poisbra!
GTPSA.mad_tpsa_taylor!
GTPSA.mad_tpsa_axpb!
GTPSA.mad_tpsa_axpbypc!
GTPSA.mad_tpsa_axypb!
GTPSA.mad_tpsa_axypbzpc!
GTPSA.mad_tpsa_axypbvwpc!
GTPSA.mad_tpsa_ax2pby2pcz2!
GTPSA.mad_tpsa_axpsqrtbpcx2!
GTPSA.mad_tpsa_logaxpsqrtbpcx2!
GTPSA.mad_tpsa_logxdy!
GTPSA.mad_tpsa_vec2fld!
GTPSA.mad_tpsa_fld2vec!
GTPSA.mad_tpsa_fgrad!
GTPSA.mad_tpsa_liebra!
GTPSA.mad_tpsa_exppb!
GTPSA.mad_tpsa_logpb!
GTPSA.mad_tpsa_mnrm
GTPSA.mad_tpsa_minv!
GTPSA.mad_tpsa_pminv!
GTPSA.mad_tpsa_compose!
GTPSA.mad_tpsa_translate!
GTPSA.mad_tpsa_eval!
GTPSA.mad_tpsa_mconv!
GTPSA.mad_tpsa_print
GTPSA.mad_tpsa_scan
GTPSA.mad_tpsa_scan_hdr
GTPSA.mad_tpsa_scan_coef!
GTPSA.mad_tpsa_debug
GTPSA.mad_tpsa_isvalid
GTPSA.mad_tpsa_init!
```
# CTPSA 
```@docs
GTPSA.CTPSA
GTPSA.mad_ctpsa_newd
GTPSA.mad_ctpsa_new
GTPSA.mad_ctpsa_del!
GTPSA.mad_ctpsa_desc
GTPSA.mad_ctpsa_uid!
GTPSA.mad_ctpsa_len
GTPSA.mad_ctpsa_nam
GTPSA.mad_ctpsa_ord
GTPSA.mad_ctpsa_ordv
GTPSA.mad_ctpsa_ordn
GTPSA.mad_ctpsa_copy!
GTPSA.mad_ctpsa_sclord!
GTPSA.mad_ctpsa_getord!
GTPSA.mad_ctpsa_cutord!
GTPSA.mad_ctpsa_maxord
GTPSA.mad_ctpsa_convert!
GTPSA.mad_ctpsa_setvar!
GTPSA.mad_ctpsa_setvar_r!
GTPSA.mad_ctpsa_setval!
GTPSA.mad_ctpsa_setval_r!
GTPSA.mad_ctpsa_setnam!
GTPSA.mad_ctpsa_clear!
GTPSA.mad_ctpsa_isnul
GTPSA.mad_ctpsa_cplx!
GTPSA.mad_ctpsa_real!
GTPSA.mad_ctpsa_imag!
GTPSA.mad_ctpsa_cabs!
GTPSA.mad_ctpsa_carg!
GTPSA.mad_ctpsa_unit!
GTPSA.mad_ctpsa_rect!
GTPSA.mad_ctpsa_polar!
GTPSA.mad_ctpsa_mono!
GTPSA.mad_ctpsa_idxs
GTPSA.mad_ctpsa_idxm
GTPSA.mad_ctpsa_idxsm
GTPSA.mad_ctpsa_cycle!
GTPSA.mad_ctpsa_get0
GTPSA.mad_ctpsa_geti
GTPSA.mad_ctpsa_gets
GTPSA.mad_ctpsa_getm
GTPSA.mad_ctpsa_getsm
GTPSA.mad_ctpsa_set0!
GTPSA.mad_ctpsa_seti!
GTPSA.mad_ctpsa_sets!
GTPSA.mad_ctpsa_setm!
GTPSA.mad_ctpsa_setsm!
GTPSA.mad_ctpsa_get0_r!
GTPSA.mad_ctpsa_geti_r!
GTPSA.mad_ctpsa_gets_r!
GTPSA.mad_ctpsa_getm_r!
GTPSA.mad_ctpsa_getsm_r!
GTPSA.mad_ctpsa_set0_r!
GTPSA.mad_ctpsa_seti_r!
GTPSA.mad_ctpsa_sets_r!
GTPSA.mad_ctpsa_setm_r!
GTPSA.mad_ctpsa_setsm_r!
GTPSA.mad_ctpsa_getv!
GTPSA.mad_ctpsa_setv!
GTPSA.mad_ctpsa_equ
GTPSA.mad_ctpsa_dif!
GTPSA.mad_ctpsa_add!
GTPSA.mad_ctpsa_sub!
GTPSA.mad_ctpsa_mul!
GTPSA.mad_ctpsa_div!
GTPSA.mad_ctpsa_pow!
GTPSA.mad_ctpsa_powi!
GTPSA.mad_ctpsa_pown!
GTPSA.mad_ctpsa_pown_r!
GTPSA.mad_ctpsa_equt
GTPSA.mad_ctpsa_dift!
GTPSA.mad_ctpsa_tdif!
GTPSA.mad_ctpsa_addt!
GTPSA.mad_ctpsa_subt!
GTPSA.mad_ctpsa_tsub!
GTPSA.mad_ctpsa_mult!
GTPSA.mad_ctpsa_divt!
GTPSA.mad_ctpsa_tdiv!
GTPSA.mad_ctpsa_powt!
GTPSA.mad_ctpsa_tpow!
GTPSA.mad_ctpsa_nrm
GTPSA.mad_ctpsa_conj!
GTPSA.mad_ctpsa_sqrt!
GTPSA.mad_ctpsa_exp!
GTPSA.mad_ctpsa_log!
GTPSA.mad_ctpsa_sincos!
GTPSA.mad_ctpsa_sin!
GTPSA.mad_ctpsa_cos!
GTPSA.mad_ctpsa_tan!
GTPSA.mad_ctpsa_cot!
GTPSA.mad_ctpsa_sinc!
GTPSA.mad_ctpsa_sincosh!
GTPSA.mad_ctpsa_sinh!
GTPSA.mad_ctpsa_cosh!
GTPSA.mad_ctpsa_tanh!
GTPSA.mad_ctpsa_coth!
GTPSA.mad_ctpsa_sinhc!
GTPSA.mad_ctpsa_asin!
GTPSA.mad_ctpsa_acos!
GTPSA.mad_ctpsa_atan!
GTPSA.mad_ctpsa_acot!
GTPSA.mad_ctpsa_asinc!
GTPSA.mad_ctpsa_asinh!
GTPSA.mad_ctpsa_acosh!
GTPSA.mad_ctpsa_atanh!
GTPSA.mad_ctpsa_acoth!
GTPSA.mad_ctpsa_asinhc!
GTPSA.mad_ctpsa_erf!
GTPSA.mad_ctpsa_erfc!
GTPSA.mad_ctpsa_acc!
GTPSA.mad_ctpsa_scl!
GTPSA.mad_ctpsa_inv!
GTPSA.mad_ctpsa_invsqrt!
GTPSA.mad_ctpsa_hypot!
GTPSA.mad_ctpsa_hypot3!
GTPSA.mad_ctpsa_integ!
GTPSA.mad_ctpsa_deriv!
GTPSA.mad_ctpsa_derivm!
GTPSA.mad_ctpsa_poisbra!
GTPSA.mad_ctpsa_taylor!
GTPSA.mad_ctpsa_poisbrat!
GTPSA.mad_ctpsa_tpoisbra!
GTPSA.mad_ctpsa_acc_r!
GTPSA.mad_ctpsa_scl_r!
GTPSA.mad_ctpsa_inv_r!
GTPSA.mad_ctpsa_invsqrt_r!
GTPSA.mad_ctpsa_axpb!
GTPSA.mad_ctpsa_axpbypc!
GTPSA.mad_ctpsa_axypb!
GTPSA.mad_ctpsa_axypbzpc!
GTPSA.mad_ctpsa_axypbvwpc!
GTPSA.mad_ctpsa_ax2pby2pcz2!
GTPSA.mad_ctpsa_axpsqrtbpcx2!
GTPSA.mad_ctpsa_logaxpsqrtbpcx2!
GTPSA.mad_ctpsa_logxdy!
GTPSA.mad_ctpsa_axpb_r!
GTPSA.mad_ctpsa_axpbypc_r!
GTPSA.mad_ctpsa_axypb_r!
GTPSA.mad_ctpsa_axypbzpc_r!
GTPSA.mad_ctpsa_axypbvwpc_r!
GTPSA.mad_ctpsa_ax2pby2pcz2_r!
GTPSA.mad_ctpsa_axpsqrtbpcx2_r!
GTPSA.mad_ctpsa_logaxpsqrtbpcx2_r!
GTPSA.mad_ctpsa_vec2fld!
GTPSA.mad_ctpsa_fld2vec!
GTPSA.mad_ctpsa_fgrad!
GTPSA.mad_ctpsa_liebra!
GTPSA.mad_ctpsa_exppb!
GTPSA.mad_ctpsa_logpb!
GTPSA.mad_ctpsa_mnrm
GTPSA.mad_ctpsa_minv!
GTPSA.mad_ctpsa_pminv!
GTPSA.mad_ctpsa_compose!
GTPSA.mad_ctpsa_translate!
GTPSA.mad_ctpsa_eval!
GTPSA.mad_ctpsa_mconv!
GTPSA.mad_ctpsa_print
GTPSA.mad_ctpsa_scan
GTPSA.mad_ctpsa_scan_hdr
GTPSA.mad_ctpsa_scan_coef!
GTPSA.mad_ctpsa_debug
GTPSA.mad_ctpsa_isvalid
GTPSA.mad_ctpsa_init!
```
